@@ -1,5 +1,5 @@
 # util/langhelpers.py
-# Copyright (C) 2005-2017 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -9,8 +9,8 @@
 modules, classes, hierarchies, attributes, functions, and methods.
 
 """
-import itertools
 import inspect
+import itertools
 import operator
 import re
 import sys
@@ -49,6 +49,11 @@ class safe_reraise(object):
 
     """
 
+    __slots__ = ('warn_only', '_exc_info')
+
+    def __init__(self, warn_only=False):
+        self.warn_only = warn_only
+
     def __enter__(self):
         self._exc_info = sys.exc_info()
 
@@ -57,7 +62,8 @@ class safe_reraise(object):
         if type_ is None:
             exc_type, exc_value, exc_tb = self._exc_info
             self._exc_info = None   # remove potential circular references
-            compat.reraise(exc_type, exc_value, exc_tb)
+            if not self.warn_only:
+                compat.reraise(exc_type, exc_value, exc_tb)
         else:
             if not compat.py3k and self._exc_info and self._exc_info[1]:
                 # emulate Py3K's behavior of telling us when an exception
@@ -187,6 +193,9 @@ class PluginLoader(object):
         self.impls = {}
         self.auto_fn = auto_fn
 
+    def clear(self):
+        self.impls.clear()
+
     def load(self, name):
         if name in self.impls:
             return self.impls[name]()
@@ -271,7 +280,7 @@ try:
 
 except ImportError:
     def inspect_func_args(fn):
-        names, _, has_kw, _ = inspect.getargspec(fn)
+        names, _, has_kw, _ = compat.inspect_getargspec(fn)
         return names, bool(has_kw)
 
 
@@ -362,7 +371,7 @@ def format_argspec_plus(fn, grouped=True):
     else:
         # we accept an existing argspec...
         spec = fn
-    args = inspect.formatargspec(*spec)
+    args = compat.inspect_formatargspec(*spec)
     if spec[0]:
         self_arg = spec[0][0]
     elif spec[1]:
@@ -371,8 +380,8 @@ def format_argspec_plus(fn, grouped=True):
         self_arg = None
 
     if compat.py3k:
-        apply_pos = inspect.formatargspec(spec[0], spec[1],
-                                          spec[2], None, spec[4])
+        apply_pos = compat.inspect_formatargspec(
+            spec[0], spec[1], spec[2], None, spec[4])
         num_defaults = 0
         if spec[3]:
             num_defaults += len(spec[3])
@@ -380,7 +389,7 @@ def format_argspec_plus(fn, grouped=True):
             num_defaults += len(spec[4])
         name_args = spec[0] + spec[4]
     else:
-        apply_pos = inspect.formatargspec(spec[0], spec[1], spec[2])
+        apply_pos = compat.inspect_formatargspec(spec[0], spec[1], spec[2])
         num_defaults = 0
         if spec[3]:
             num_defaults += len(spec[3])
@@ -391,9 +400,9 @@ def format_argspec_plus(fn, grouped=True):
     else:
         defaulted_vals = ()
 
-    apply_kw = inspect.formatargspec(name_args, spec[1], spec[2],
-                                     defaulted_vals,
-                                     formatvalue=lambda x: '=' + x)
+    apply_kw = compat.inspect_formatargspec(
+        name_args, spec[1], spec[2], defaulted_vals,
+        formatvalue=lambda x: '=' + x)
     if grouped:
         return dict(args=args, self_arg=self_arg,
                     apply_pos=apply_pos, apply_kw=apply_kw)
@@ -529,20 +538,24 @@ class portable_instancemethod(object):
 
     """
 
-    __slots__ = 'target', 'name', '__weakref__'
+    __slots__ = 'target', 'name', 'kwargs', '__weakref__'
 
     def __getstate__(self):
-        return {'target': self.target, 'name': self.name}
+        return {'target': self.target, 'name': self.name,
+                'kwargs': self.kwargs}
 
     def __setstate__(self, state):
         self.target = state['target']
         self.name = state['name']
+        self.kwargs = state.get('kwargs', ())
 
-    def __init__(self, meth):
+    def __init__(self, meth, kwargs=()):
         self.target = meth.__self__
         self.name = meth.__name__
+        self.kwargs = kwargs
 
     def __call__(self, *arg, **kw):
+        kw.update(self.kwargs)
         return getattr(self.target, self.name)(*arg, **kw)
 
 
@@ -563,7 +576,7 @@ def class_hierarchy(cls):
         if isinstance(cls, types.ClassType):
             return list()
 
-    hier = set([cls])
+    hier = {cls}
     process = list(cls.__mro__)
     while process:
         c = process.pop()
@@ -633,8 +646,8 @@ def monkeypatch_proxied_specials(into_cls, from_cls, skip=None, only=None,
             continue
         try:
             spec = compat.inspect_getargspec(fn)
-            fn_args = inspect.formatargspec(spec[0])
-            d_args = inspect.formatargspec(spec[0][1:])
+            fn_args = compat.inspect_formatargspec(spec[0])
+            d_args = compat.inspect_formatargspec(spec[0][1:])
         except TypeError:
             fn_args = '(self, *args, **kw)'
             d_args = '(*args, **kw)'
@@ -846,7 +859,7 @@ def dependency_for(modulename):
         # unfortunately importlib doesn't work that great either
         tokens = modulename.split(".")
         mod = compat.import_(
-            ".".join(tokens[0:-1]), globals(), locals(), tokens[-1])
+            ".".join(tokens[0:-1]), globals(), locals(), [tokens[-1]])
         mod = getattr(mod, tokens[-1])
         setattr(mod, obj.__name__, obj)
         return obj
